#include <smmintrin.h>
#include <immintrin.h>
#include <stdalign.h>
#include <stdio.h>
#include <stdint.h>

uint8_t kb[] = { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
             0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f };

uint8_t nb[] = { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
             0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x1f };

uint8_t pb[] = { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
             0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x2f };

extern __m128i enc(__m128i k, __m128i n,__m128i p);
extern __m128i dec(__m128i k, __m128i n,__m128i c);

extern void enc_mem(uint8_t *cp, uint8_t *kp, uint8_t *np, uint8_t *pp);
extern void dec_mem(uint8_t *pp, uint8_t *kp, uint8_t *np, uint8_t *cp);

int crypto_verify(const uint8_t *x,const uint8_t *y)
{
  unsigned int differentbits = 0;
#define F(i) differentbits |= x[i] ^ y[i];
  F(0)
    F(1)
    F(2)
    F(3)
    F(4)
    F(5)
    F(6)
    F(7)
    F(8)
    F(9)
    F(10)
    F(11)
    F(12)
    F(13)
    F(14)
    F(15)
    return (1 & ((differentbits - 1) >> 8)) - 1;
}

void
test_simd_interface(void) {
  __m128i c, pp, p, k, n;

  k = _mm_loadu_si128((__m128i *) kb);
  n = _mm_loadu_si128((__m128i *) nb);
  p = _mm_loadu_si128((__m128i *) pb);

  c = enc(k,n,p);
  pp = dec(k,n,c);

  __m128i neq = _mm_xor_si128(p,pp);
  if(_mm_test_all_zeros(neq, neq)) {
    printf("Verify output: OK!\n");
  }
  else {
    printf("Verify output: Not OK!\n");
  }
}

void
test_mem_interface(void) {
  uint8_t c[16], pp[16];

  enc_mem(c, kb, nb, pb);
  dec_mem(pp, kb, nb, c);

  if (crypto_verify(pb, pp) == 0) {
    printf("Verify output: OK!\n");
  }
  else {
    printf("Verify output: Not OK!\n");
  }

}

int
main(void) {
  test_simd_interface();
  test_mem_interface();
  return 0;
}
